# -*- ruby -*-

#
# VFS for ruby - open anything with URI
# 
#   Copyright: Tatsuki Sugiura <sugi@nemui.org>
#   License: Ruby's original or GPL2
#   $LastChangedDate: 2003-11-20 09:16:53 +0900 (Thu, 20 Nov 2003) $
#   $LastChangedBy: sugi $
#

require 'net/http'
require 'thread'
#if ::VERSION.split(/\./)[0..1].join(".").to_f > 1.7
#  Net::HTTP.version_1_2
#end
Net::HTTP.version_1_1
unless Net::HTTP.instance_methods(true).member? "propfind"
  require 'vfs/scheme/http/davext'
end

# FIXME!! will be changed.
unless Net::HTTPGenericRequest.constants.member? "SEND_BUFFER_SIZE"
  require 'vfs/scheme/http/streamext'
end

module VFS;  module SchemeHandler;  module HTTP;
  class ConnectPool
    @@pool = Hash.new
    CALLER = 0
    CONN   = 1
    MUTEX  = 3
    class << self
      def connect(caller_id, host, port)
        host_id = host + ":" + port.to_s
        if ! @@pool.has_key?(host_id)
          if ENV.member? "http_proxy"
            proxy = URI.parse(ENV["http_proxy"])
            proxy_host = proxy.host
            proxy_port = proxy.port
          end
          @@pool[host_id] = Array.new
          @@pool[host_id][CONN] = Net::HTTP.new(host, port,
                                                proxy_host,
                                                proxy_port)
          log("Opening HTTP connection to #{host_id}.", Log::INFO)
          @@pool[host_id][CONN].start
          @@pool[host_id][CALLER] = [caller_id]
          @@pool[host_id][MUTEX] = Mutex.new
        elsif @@pool[host_id][CALLER].member? caller_id
          @@pool[host_id][CALLER] << caller_id
        end

        if block_given?
          yield(@@pool[host_id][CONN], @@pool[host_id][MUTEX])
        else
          [@@pool[host_id][CONN], @@pool[host_id][MUTEX]]
        end
      end

      def request(caller_id, uri, method=:get, *rest_args, &block)
        t_uri = uri
        begin
          connect(caller_id, t_uri.host, t_uri.port) { |http, mutex|
            log("Requesting '#{method}' for '#{t_uri.path}' to server.",
                Log::DEBUG)
            response = nil
            mutex.synchronize {
              response = http.send(method, t_uri.path, *rest_args)
            }
            Array === response and response = response[0]
            log("Data was fetched from web server. Go yield.",
                Log::DEBUG)
            if block_given?
              yield(t_uri, response)
            else
              [t_uri, response]
            end
          }
        rescue Net::ProtoRetriableError => err
          log("Retriable error was occured in HTTP request. (detail: #{err})",
              Log::INFO)
          t_uri = URI.parse(err.response['location'])
          retry
        end
      end

      def release(caller_id)
        @@pool.each { |hostid, pool|
          pool[CALLER].delete(caller_id)
          if pool[CALLER].empty?
            if pool[CONN].active?
              log("Closing connection for #{hostid}.", Log::INFO)
              pool[CONN].finish
            end
             @@pool.delete hostid
          end
        }
      end

      def log(msg, level)
        VFS::Log.log("VFS(HTTP): #{msg}", level)
      end
    end
  end

  module Common

    PROP_CACHE_TIME = 30

    def check_scheme(uri)
      unless uri.scheme == "http"
	raise Error::InvalidURI,
	  "Invalid HTTP URL '#{uri}' was speficied."
      end
    end

    def init_common_attr
      @uri.path = "/" if @uri.path == ""
      @capability   = []
      @prop_cache   = nil
      @prop_cache_t = 0
      @header       = nil
      ObjectSpace.define_finalizer(self,
                                   Proc.new { |id| 
          Proc.new { ConnectPool.release(id) }
      }.call(self.__id__))
    end
    private :init_common_attr
    attr_writer :capability, :header
    attr_accessor :prop_cache, :prop_cache_t
    private :prop_cache, :prop_cache=, :prop_cache_t,
      :prop_cache_t=, :header=

    def close_socket
      VFS::SchemeHandler::HTTP::Common.cleanup(@conn)
    end

    def capability?(method)
      capability.member? method      
    end

    def capability
      return @capability unless @capability.length == 0
      begin
        capability = http_request(:options) { |uri, rep|
          rep["allow"].split(/\s*,\s*/)
        }
      rescue Net::HTTPServerException, Net::HTTPFatalError
        capability = ["GET"]
      end
    end

    def property(elem)
      unless capability? "PROPFIND"
	return nil
      end
      if prop_cache == nil || Time.now.to_i - prop_cache_t > PROP_CACHE_TIME
	log("Start fetching property.")
        begin
          http_request(:propfind, nil, {"Depth" => "0"}) { |uri, rep|
            log("*BUG/NotInplementedYet* must unescape XML entity.", Log::WARNING)
            __send__ :prop_cache=, rep.body
          }
        rescue Net::HTTPServerException
          return nil
        end
	__send__ :prop_cache_t=, Time.now.to_i
	log("Property was fetched")
      end
      log("Scaning property '#{elem}'.", Log::DEBUG)
      m = Regexp.new("<#{elem}(?:\s+[^>]*)?>(.*?)</#{elem}>").match(prop_cache)
      if m && m[1]
	m[1].tr("\n\r\t", " ")
      else
	nil
      end
    end

    def header(name=nil)
      if @header == nil
	log("Start fetching header.")
	http_request(:head) { |uri, rep|
	  __send__ :header=, rep.each_header{}
	}
	log("Header was fetched")
      end
      name == nil ? @header : @header[name]
    end

    def http_request(method=:get, *rest_args, &block)
      ConnectPool.request(self.__id__, uri, method, *rest_args, &block)
    end
  end # Common
end; end; end
